@@ -1,9 +1,11 @@
+import datetime
 from functools import wraps
 import importlib.resources
 import inspect
 import json
 from typing import Callable, Optional
 from dataclasses import dataclass, field
+from numpy import int64
 
 import pandas
 import requests
@@ -22,10 +24,17 @@ class Expectation:
     args: dict = field(default_factory=dict)
     result: dict = field(default_factory=dict)
     parent: "DataframeTest" = None
+    failed_sample_indices: Optional[list] = None
+    failed_sample: Optional[pandas.DataFrame] = None
 
     @staticmethod
-    def Fail(message: str, **kwargs):
-        return Expectation(success=False, message=message, result=kwargs)
+    def Fail(message: str, failed_sample_indices=None, **kwargs):
+        return Expectation(
+            success=False,
+            message=message,
+            failed_sample_indices=failed_sample_indices,
+            result=kwargs,
+        )
 
     @staticmethod
     def Pass(message: Optional[str] = None, **kwargs):
@@ -39,6 +48,14 @@ class Expectation:
             "message": self.message,
             "args": self.args,
             "result": self.result,
+            "failed_sample_indices": [
+                coerce_to_json(i) for i in self.failed_sample_indices
+            ]
+            if self.failed_sample_indices
+            else None,
+            "failed_sample": self.failed_sample.to_dict(orient="records")
+            if self.failed_sample is not None
+            else None,
         }
 
     def _repr_markdown_(self):
@@ -124,11 +141,12 @@ class SeriesTest:
             self.unit = unit
         return self
 
-    def __enter__(self):
-        return self
-
-    def __exit__(self, _type, _value, _traceback):
-        return self
+def coerce_to_json(arg):
+    if inspect.isfunction(arg):
+        return arg.__name__
+    if isinstance(arg, int64):
+        return int(arg)
+    return arg
 
 
 def expectation(func):
@@ -140,11 +158,25 @@ def expectation(func):
         result.critical = self.critical
         result.parent = self.parent.parent
 
-        args_names = list(inspect.signature(func).parameters.keys())[1:]
-        result.args = {
-            **dict(zip(args_names, args)),
-            **kwargs,
-        }
+        result.args = {**kwargs}
+        for i, parameter in enumerate(inspect.signature(func).parameters.values()):
+            if parameter.name == "self":
+                continue
+            if parameter.kind in [
+                inspect.Parameter.VAR_KEYWORD,
+                inspect.Parameter.VAR_POSITIONAL,
+            ]:
+                result.args[parameter.name] = [coerce_to_json(arg) for arg in args]
+            else:
+                result.args[parameter.name] = coerce_to_json(
+                    args[i - 1] if len(args) >= i else parameter.default
+                )
+
+        if result.failed_sample_indices and not result.failed_sample:
+            if type(self.parent.parent.dataframe) == pandas.DataFrame:
+                result.failed_sample = self.parent.parent.dataframe.filter(
+                    items=result.failed_sample_indices, axis=0
+                )
 
         self.expectations.append(result)
         return result
@@ -160,6 +192,10 @@ class Asserter:
         self.expectations: "list[Expectation]" = []
 
     def bins(self):
+        if type(self.series) is not pandas.Series:
+            # TODO: implement for dask
+            return None
+
         bins = pandas.cut(self.series, bins=10).value_counts()
         return json.loads(bins.to_json())
 
@@ -218,16 +254,22 @@ class Asserter:
             return Expectation.Fail(
                 f"out of range. expected to be in ({minimum}, {maximum}) \n"
                 + f"but found ({found_min}, {found_max}).",
+                failed_sample_indices=[
+                    compute(self.series.idxmin()),
+                    compute(self.series.idxmax()),
+                ],
                 **result,
             )
         elif extends_left:
             return Expectation.Fail(
                 f"expected values to be at least {minimum}, but found {found_min}",
+                failed_sample_indices=[compute(self.series.idxmin())],
                 **result,
             )
         elif extends_right:
             return Expectation.Fail(
                 f"expected values to be at most {maximum}, but found {found_max}",
+                failed_sample_indices=[compute(self.series.idxmax())],
                 **result,
             )
 
@@ -248,7 +290,9 @@ class Asserter:
 
         if found_min < 0:
             return Expectation.Fail(
-                f"negative value found: min is {found_min}", **result
+                f"negative value found: min is {found_min}",
+                failed_sample_indices=[compute(self.series.idxmin())],
+                **result,
             )
 
         return Expectation.Pass(**result)
@@ -268,7 +312,9 @@ class Asserter:
 
         if found_max < 0:
             return Expectation.Fail(
-                f"positive value found: max is {found_max}", **result
+                f"positive value found: max is {found_max}",
+                failed_sample_indices=[compute(self.series.idxmax())],
+                **result,
             )
 
         return Expectation.Pass(**result)
@@ -276,28 +322,49 @@ class Asserter:
     @expectation
     def not_be_null(self):
         """
-        checks if there are any null values.
+        checks if all values are non-null.
 
         Examples:
             >>> dp.user_id.must.not_be_null()
         """
 
-        if self.series.isnull().values.any():
-            return Expectation.Fail("found null values")
+        null_index = self.series.isnull().first_valid_index()
+
+        if null_index:
+            return Expectation.Fail(
+                "found null values",
+                failed_sample_indices=[null_index],
+            )
 
         return Expectation.Pass()
 
     @expectation
-    def be_one_of(self, *args):
+    def not_be_na(self):
         """
-        checks if there's any value not in the given list.
+        checks if all values are non-na.
+
+        Examples:
+            >>> dp.user_id.must.not_be_na()
+        """
+
+        na_index = self.series.isna().first_valid_index()
+
+        if na_index:
+            return Expectation.Fail("found na values", failed_sample_indices=[na_index])
+
+        return Expectation.Pass()
+
+    @expectation
+    def be_one_of(self, *allowed_values):
+        """
+        checks if values are in the allowed values.
 
         Examples:
             >>> dp.state.must.be_one_of("active", "sleeping", "inactive")
         """
 
         existing = set(compute(self.series.unique()))
-        expected = set(args)
+        expected = set(allowed_values)
 
         additional = existing - expected
 
@@ -306,8 +373,13 @@ class Asserter:
         if len(additional) > 0:
             additional_values = list(additional)
             additional_values.sort()
+
             return Expectation.Fail(
                 f"found additional values: {additional_values}",
+                failed_sample_indices=[
+                    compute(self.series.eq(value)).idxmax()
+                    for value in additional_values[:5]
+                ],
                 used_pct=used_pct,
             )
 
@@ -319,12 +391,19 @@ class Asserter:
         checks if all values are ISO8601-compliant dates.
         datetimes will be rejected.
 
-        TODO: implement
-
         Examples:
             >>> dp.day.must.be_date()
         """
 
+        rejected = compute(
+            self.series[-self.series.str.match(r"\d{4}-\d{2}-\d{2}")].index
+        ).to_list()
+
+        if len(rejected) > 0:
+            return Expectation.Fail(
+                "found non-date values", failed_sample_indices=rejected[:5]
+            )
+
         return Expectation.Pass()
 
     @expectation
@@ -332,12 +411,19 @@ class Asserter:
         """
         checks if all values are ISO8601-compliant datetimes.
 
-        TODO: implement
-
         Examples:
             >>> dp.timestamp.must.be_datetime()
         """
 
+        rejected = compute(
+            self.series[-pandas.to_datetime(self.series, errors="coerce").isna()].index
+        ).to_list()
+
+        if len(rejected) > 0:
+            return Expectation.Fail(
+                "found non-datetime values", failed_sample_indices=rejected[:5]
+            )
+
         return Expectation.Pass()
 
     @expectation
@@ -345,13 +431,34 @@ class Asserter:
         """
         checks if all values are unix epoch-compliant timestamps.
 
-        TODO: implement
-
         Examples:
             >>> dp.timestamp.must.be_unix_epoch()
         """
 
-        return Expectation.Pass()
+        found_min = compute(self.series.min())
+        found_max = compute(self.series.max())
+
+        result = {
+            "minimum": found_min,
+            "maximum": found_max,
+        }
+
+        if found_min < 0:
+            return Expectation.Fail(
+                f"unix epoch times should be positive. found {found_min}, which is before 1970.",
+                failed_sample_indices=[compute(self.series.idxmin())],
+                **result,
+            )
+
+        max_timestamp = datetime.datetime(2100, 0, 0).timestamp()
+        if found_max > max_timestamp:
+            return Expectation.Fail(
+                f"found {found_max}, which is after the year 2100",
+                failed_sample_indices=[compute(self.series.idxmax())],
+                **result,
+            )
+
+        return Expectation.Pass(**result)
 
     @expectation
     def fulfill(self, custom_assertion: Callable[[pandas.Series], Optional[str]]):
@@ -433,12 +540,6 @@ class DataframeTest:
     def __getitem__(self, key):
         return self.__getattr__(key)
 
-    def __enter__(self):
-        return self
-
-    def __exit__(self, _type, _value, _traceback):
-        return self
-
     def collect(self) -> DataframeResult:
         """
         Collects all tests and returns a DataframeResult object.
@@ -462,6 +563,46 @@ class DataframeTest:
 
         return result
 
+    def expectations(self) -> "list[Expectation]":
+        """
+        Returns a list of all executed expectations.
+        """
+        result = []
+        for s in self.collect().series:
+            result += s.expectations
+        return result
+
+    def failed_expectations(self) -> Expectation:
+        """
+        Returns a list of all failed expectations.
+        """
+        return [e for e in self.expectations() if not e.success]
+
+    def failed_critical_expectations(self) -> Expectation:
+        """
+        Returns a list of all failed critical expectations.
+        """
+        return [e for e in self.failed_expectations() if e.critical]
+
+    def is_failure(self) -> bool:
+        """
+        Returns True if one of the expectations failed.
+        """
+        return len(self.failed_expectations()) > 0
+
+    def is_critical_failure(self) -> bool:
+        """
+        Returns True if a critical expectation failed.
+        """
+        return len(self.failed_critical_expectations()) > 0
+
+    def check(self):
+        """
+        Raises an exception if a critical expectation failed.
+        """
+        if self.is_critical_failure():
+            raise Exception("critical expectation failed")
+
     def _repr_markdown_(self):
         """
         see https://ipython.readthedocs.io/en/stable/config/integrating.html#rich-display
@@ -472,11 +613,11 @@ class DataframeTest:
 
         for series in result.series:
             md += f"**{series.name}**  \n"
-            for expectation in series.expectations:
-                if expectation.success:
-                    md += f"✅ {expectation.name}  \n"
+            for e in series.expectations:
+                if e.success:
+                    md += f"✅ {e.name}  \n"
                 else:
-                    md += f"❌ {expectation.name}: {expectation.message}  \n"
+                    md += f"❌ {e.name}: {e.message}  \n"
             md += "\n"
 
         return md
